#include <upcxx/future/core.hpp>

using namespace std;

using upcxx::detail::future_header;
using upcxx::detail::future_header_nil;
using upcxx::detail::future_header_dependent;
using upcxx::detail::future_body;
using upcxx::detail::future_body_proxy_;

template<typename ...T>
using future_header_result = upcxx::detail::future_header_result<T...>;

constexpr future_header future_header_nil::the_nil
#if 0
= {
  /*ref_n_*/-1,
  /*status_*/future_header::status_active + 666,
  /*sucs_head_*/nullptr,
  {/*result_*/nullptr}
}
#endif
;

future_header future_header_result<>::the_always = {
  /*ref_n_*/-1,
  /*status_*/upcxx::detail::future_header::status_ready,
  /*sucs_head_*/nullptr,
  {/*result_*/&upcxx::detail::future_header_result<>::the_always}
};

namespace {
  __thread future_header_dependent **active_tail_ = nullptr;
}

void future_header::dependency_link::unlink() {
  if(suc != nullptr) {
    future_header::dependency_link **pp = &dep->sucs_head_;
    while(*pp != this)
      pp = &(*pp)->sucs_next;
    *pp = (*pp)->sucs_next;
  }
}

void future_header_dependent::entered_active() {
  this->incref(1); // being in active queue is a reference
  
  if(active_tail_ != nullptr) {
    // add to existing active queue
    *active_tail_ = this;
    this->active_next_ = nullptr;
    active_tail_ = &this->active_next_;
  }
  else { // no active queue exists
    future_header_dependent *active_head = nullptr;
    active_tail_ = &active_head;
    
    // we are only inhabitant of active queue
    this->body_->leave_active(this);
    
    // active queue may have been just appended to, deplete it
    future_header_dependent *cur = active_head;
    while(cur != nullptr) {
      future_header_dependent *next = cur->active_next_;
      if(next == nullptr) {
        active_head = nullptr;
        active_tail_ = &active_head;
      }
      cur->body_->leave_active(cur);
      cur = next ? next : active_head;
    }
    
    // dislodge our queue
    active_tail_ = nullptr;
  }
}

void future_header::entered_ready_with_sucs(future_header *result, dependency_link *sucs_head) {
  /* Outlined:
  // caller gave us a reference in result->ref_n_
  this->result_ = result;
  this->status_ = future_header::status_ready;
  */
  
  // keep track of references we lose as we give our result to our successors
  int this_refs = this->ref_n_;
  UPCXX_ASSERT(this_refs > 0);
  int this_refs_unit = result == this ? 0 : 1;
  
  int result_refs = result->ref_n_;
  int result_refs_unit = result == this || result_refs < 0 ? 0 : 1;
  
  // begin an active queue if one isnt already in place
  future_header_dependent *active_head = nullptr;
  bool our_active_queue = active_tail_ == nullptr;
  if(our_active_queue)
    active_tail_ = &active_head;
  
  { // nobody will enter our sucs list while we traverse, because:
    // 1. we defer draining til after traversal, no functions called here.
    // 2. we're ready, so nobody should be entering our sucs_head_ anyway
    dependency_link *link = sucs_head;
    this->sucs_head_ = nullptr;
    
    while(link != nullptr) {
      future_header_dependent *suc = link->suc;
      
      link->dep = result;
      link->suc = nullptr; // mark as unlinked
      
      this_refs -= this_refs_unit;
      result_refs += result_refs_unit;
      
      if(--suc->status_ <= future_header::status_active) {
        suc->incref(1); // being in active queue is a reference
        // add to active queue
        *active_tail_ = suc;
        suc->active_next_ = nullptr;
        active_tail_ = &suc->active_next_;
      }
      
      link = link->sucs_next;
    }
  }
  
  if(0 == this_refs)
    // we're dying so take one back from result.
    // only possible if we've lost references, hence result_refs will not cross zero.
    result_refs -= result_refs_unit;
  
  // write back this->ref_n_
  {int trash; (this_refs_unit == 1 ? this->ref_n_ : trash) = this_refs;}
  // write back result->ref_n_
  {int trash; (result_refs_unit == 1 ? result->ref_n_ : trash) = result_refs;}
  
  if(0 == this_refs) {
    // we know result != this, hence we're a future_header_dependent
    delete static_cast<future_header_dependent*>(this);  
  }
  
  // if we made an active queue then we have to deplete it
  if(our_active_queue) {
    future_header_dependent *cur = active_head;
    while(cur != nullptr) {
      future_header_dependent *next = cur->active_next_;
      if(next == nullptr) {
        active_head = nullptr;
        active_tail_ = &active_head;
      }
      cur->body_->leave_active(cur);
      cur = next ? next : active_head;
    }
    active_tail_ = nullptr;
  }
}

void future_header_dependent::enter_proxying(
    future_body_proxy_ *body,
    future_header *proxied
  ) {
  
  // we defer deleting memory until the end to help compiler keep things in registers
  void *deferred_delete_1 = nullptr;
  future_header_dependent *deferred_delete_2 = nullptr;
  void *deferred_delete_3 = nullptr;
  
  if(proxied->status_ == future_header::status_proxying ||
     proxied->status_ == future_header::status_proxying_active
    ) {
    future_body_proxy_ *proxied_body = static_cast<future_body_proxy_*>(proxied->body_);
    future_header *proxied1 = proxied_body->link_.dep;
    
    proxied1->incref(1);
    
    if(0 == static_cast<future_header_dependent*>(proxied)->decref(1)) {
      // steal body from proxied, dont use given body
      deferred_delete_1 = body->storage_;
      body = proxied_body;
      
      // put ourself in proxied1 successor list
      body->link_.suc = this;
      
      deferred_delete_2 = static_cast<future_header_dependent*>(proxied);
    }
    
    proxied = proxied1;
  }
  
  if(proxied->status_ == future_header::status_ready) {
    deferred_delete_3 = body->storage_; // dont need body
    
    this->enter_ready(future_header::drop_for_result(proxied));
  }
  else {
    this->status_ = future_header::status_proxying;
    
    // will write back these reference counts later
    int this_refs = this->ref_n_;
    int proxied_refs = proxied->ref_n_;
    bool proxied_refs_wb = proxied_refs >= 0;
    
    // look for successors that proxy us
    for(dependency_link **plink = &this->sucs_head_; *plink != nullptr;) {
      dependency_link *link = *plink;
      future_header_dependent *suc = link->suc;
      
      if(suc->status_ == future_header::status_proxying) {
        // our successor is proxying on us, make it point past us
        static_cast<future_body_proxy_*>(suc->body_)->link_.dep = proxied;
        
        // remove link from our sucs list
        *plink = link->sucs_next;
        
        // add link to proxied sucs
        link->sucs_next = proxied->sucs_head_;
        proxied->sucs_head_ = link;
        
        // our loss is proxied's gain
        this_refs -= 1;
        proxied_refs += 1;
      }
      else
        plink = &link->sucs_next;
    }
    
    if(this_refs == 0)
      proxied_refs -= 1;
    
    // write back proxied->ref_n_
    // we know it wont be zero because either we're referencing it or our successors are
    {int trash; (proxied_refs_wb ? proxied->ref_n_ : trash) = proxied_refs;}
    
    if(this_refs == 0) {
      // nobody points to us, so we die...
      future_body::operator delete(body->storage_);
      delete this;
    }
    else {
      // point to proxied
      body->link_.dep = proxied; // give them our reference
      
      // put ourself in proxied->sucs_head_
      body->link_.suc = this;
      body->link_.sucs_next = proxied->sucs_head_;
      proxied->sucs_head_ = &body->link_;
      
      this->ref_n_ = this_refs;
      this->body_ = body;
    }
  }
  
  future_body::operator delete(deferred_delete_1);
  delete deferred_delete_2;
  future_body::operator delete(deferred_delete_3);
}

void future_body_proxy_::leave_active(future_header_dependent *hdr) {
  if(hdr->ref_n_ != 1) { // active queue is not only reference
    hdr->ref_n_ -= 1; // drop active queue reference
    
    future_header *result = this->link_.dep; // link.dep replaced with result in dependency's enter_ready()
    
    // discard the body. no destructor needed since future_body_proxy<T...>
    // is trivially destructible and we dont want to decref the proxied pointer.
    operator delete(this->storage_);
    
    hdr->enter_ready(result);
  }
  else { // only reference is active queue, just delete it
    void *storage = this->storage_;
    this->destruct_early();
    operator delete(storage);
    delete hdr;
  }
}

void future_body::destruct_early() {
  UPCXX_INVOKE_UB(); // never called
}
